import os
import sys
from typing import Union
import argparse

from py_to_ts_interfaces.enums import EnumDefinition
from py_to_ts_interfaces.file_io import write_file, read_file
from py_to_ts_interfaces.interfaces import InterfaceDefinition


def python_to_typescript_file(python_code: str) -> str:
    """
    Convert python enum and dataclass definitions to equivalent typescript code.

    :param python_code: Python code containing only enums and dataclasses.
    :return: Equivalent typescript code.
    """
    # initial processing (remove superfluous lines)
    lines = python_code.splitlines()
    lines = [line for line in lines if line and not line.isspace() and not line.startswith(("from ", "#", "@"))]

    # group the lines for each enum/class definition together
    definition_groups: list[list[str]] = []
    for line in lines:
        if line.startswith("class "):
            definition_groups.append([])
        definition_groups[-1].append(line)

    # convert each group into either an EnumDefinition or InterfaceDefinition object
    processed_definitions: list[Union[EnumDefinition, InterfaceDefinition]] = []
    for definition in definition_groups:
        if definition[0].endswith("(Enum):"):
            processed_definitions.append(EnumDefinition(definition))
        else:
            processed_definitions.append(InterfaceDefinition(definition))

    # construct final output
    typescript_output = ""
    for processed_definition in processed_definitions:
        typescript_output += "{}\n\n".format(processed_definition.get_typescript())
    typescript_output = typescript_output.strip("\n")
    # add just one newline at the end
    typescript_output += "\n"

    return typescript_output


def python_to_typescript_folder(input_path: str, output_path: str) -> None:
    """
    Convert all python files in input directory to typescript files in output directory. Each output file has the
    same name as its python source (with the file extension changed to 'ts').

    :param input_path: A full or relative path to a folder containing .py files.
    :param output_path: A full or relative path to a folder which may not exist.
    """
    for file in os.listdir(input_path):
        if file.endswith(".py") and file != "__init__.py":
            file_contents = read_file(os.path.join(input_path, file))

            typescript_output = python_to_typescript_file(file_contents)

            write_file(typescript_output, os.path.join(output_path, file[:-3] + ".ts"))


def main():
    """Main script."""
    parser = argparse.ArgumentParser()
    parser.add_argument("input_folder", help="The path to the folder of python files to be converted")
    parser.add_argument("output_folder", help="The path to the folder to output the typescript files to")
    args = parser.parse_args()

    python_to_typescript_folder(args.input_folder, args.output_folder)


if __name__ == "__main__":
    sys.exit(main())



from typing import List, Tuple

from py_to_ts_interfaces.type_converting import python_to_typescript_type
from py_to_ts_interfaces.utils import to_camel_case


class InterfaceField:
    """Represent a dataclass field."""
    name: str
    python_type: str
    is_nullable: bool

    def __init__(self, line: str):
        self.is_nullable = line.endswith(" = None") or "Union[" in line
        line = line.removesuffix(" = None")

        self.name, self.python_type = self.get_name_and_type(line)

    @staticmethod
    def get_name_and_type(line: str) -> Tuple[str, str]:
        """Take a line like "field_name: Union[None, int]" and return ("fieldName", "int")"""
        name, python_type = line.strip().split(": ")
        name = to_camel_case(name)
        if "Union[" in python_type:
            python_type = python_type.removeprefix("Union[None, ").removeprefix("Union[")
            python_type = python_type.removesuffix("]").removesuffix(", None")
        return name, python_type

    def get_typescript(self) -> str:
        """Return the field in typescript syntax (including indentation)."""
        ts_name = self.name
        if self.is_nullable:
            ts_name += "?"
        return "    {0}: {1};".format(ts_name, python_to_typescript_type(self.python_type))


class InterfaceDefinition:
    """Represent a python dataclass/typescript interface."""
    name: str
    fields: List[InterfaceField]

    def __init__(self, definition: List[str]):
        self.name = definition[0].removeprefix("class ").strip(":")
        self.fields = [InterfaceField(line) for line in definition[1:]]

    def get_typescript(self) -> str:
        """Return the entire interface in typescript syntax (including indentation)."""
        typescript_string = "export interface {0} {{\n".format(self.name)
        for field in self.fields:
            typescript_string += "{}\n".format(field.get_typescript())
        typescript_string += "}"
        return typescript_string



# Manually supporting dict types scales pretty badly, so the python_to_typescript_type function handles them
# programmatically for the sake of sanity.
python_to_typescript_type_map = {
    "str": "string",
    "int": "number",
    "float": "number",
    "complex": "number",
    "bool": "boolean",
    "List[int]": "number[]",
    "List[str]": "string[]",
}


def python_to_typescript_type(python_type: str) -> str:
    """
    Map python type to an equivalent typescript type.

    :param python_type: A python type like 'str' or 'int'.
    :return: An equivalent typescript type. If there is no known mapping for the input, then it is returned as-is for
    enum support.
    """
    try:
        return python_to_typescript_type_map[python_type]
    except KeyError:
        if python_type.startswith("Dict["):
            python_type = python_type.removeprefix("Dict[").removesuffix("]").replace(" ", "")
            # This part handles nested dicts
            py_type_1, py_type_2 = python_type.split(",", 1)
            ts_type_1 = python_to_typescript_type(py_type_1)
            ts_type_2 = python_to_typescript_type(py_type_2)
            return f"Record<{ts_type_1}, {ts_type_2}>"
        else:
            # This should mean it is an enum
            return python_type



import os


def read_file(file_path: str) -> str:
    """Read content of file provided."""
    with open(file_path, "r", encoding="utf-8") as reader:
        return reader.read()


def write_file(to_write: str, file_path: str) -> None:
    """Write input string to file."""
    folder_path = os.path.dirname(file_path)
    if folder_path and not os.path.isdir(folder_path):
        os.makedirs(folder_path)
    with open(file_path, "w+", encoding="utf-8") as writer:
        writer.write(to_write)



def to_camel_case(snake_str: str) -> str:
    """
    Convert a snake_case string to camelCase.

    :param snake_str: The input in snake_case
    :return: The input, but in camelCase
    """
    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + ''.join(x.title() for x in components[1:])
